from pathlib import Path
import cv2

from .exceptions import ElementNotFoundError


class Image:
    def __init__(self, data=None):
        self.data = data
        try:
            self.height = data.shape[0]
            self.width = data.shape[1]
        except AttributeError:
            self.height = None
            self.width = None

    @classmethod
    def load(cls, path):
        return cls(cv2.imread(str(path)))

    def save(self, path):
        path = Path(path)
        path.parent.mkdir(exist_ok=True)

        unique_path = self._find_unique_path(path)
        cv2.imwrite(str(unique_path), self.data)

        return unique_path

    def match_template(self, template, threshold=0.001):
        result = cv2.matchTemplate(self.data, template.data, cv2.TM_SQDIFF_NORMED)
        minVal, _, minLoc, _ = cv2.minMaxLoc(result)

        if minVal > threshold:
            raise ElementNotFoundError
        return (
            minLoc[0],
            minLoc[1],
            template.width,
            template.height,
        )

    @staticmethod
    def _find_unique_path(path):
        if not path.exists():
            return path

        prefix = path.with_suffix("")
        suffix = path.suffix
        count = 2
        while True:
            unique_prefix = prefix.with_name(prefix.name + f"_{count}")
            unique_path = unique_prefix.with_suffix(suffix)
            if not unique_path.exists():
                return unique_path
            count += 1



import time
from pynput.mouse import Controller, Button


class Mouse:
    def __init__(self):
        self._controller = Controller()

    def get_position(self):
        return self._controller.position

    def move_to(self, x, y):
        self._controller.position = (x, y)

    def click(self, x, y, button=Button.left):
        self.move_to(x, y)
        time.sleep(0.01)
        self._controller.click(button)



from hamcrest.core.base_matcher import BaseMatcher

EVENTUAL_TIMEOUT_SECONDS = 30


class _Visible(BaseMatcher):
    def __init__(self, timeout_seconds=0):
        self.timeout_seconds = timeout_seconds

    def _matches(self, element):
        try:
            return element.is_visible(timeout_seconds=self.timeout_seconds)
        except FileNotFoundError:
            element.save_last_screenshot()
            raise

    def describe_to(self, description):
        description.append_text("to be visible")

    def describe_mismatch(self, element, mismatch_description):
        mismatch_description.append_text("it was not found")
        element.save_last_screenshot()


def visible():
    return _Visible()


def eventually_visible(within_seconds=EVENTUAL_TIMEOUT_SECONDS):
    return _Visible(within_seconds)



from .image import Image
from .screen_grabber import ScreenGrabber


class Locater:
    def __init__(self):
        self.last_screenshot = None
        self._grabber = ScreenGrabber()

    def locate(self, image_path, boundary):
        screenshot = self._grabber.grab(boundary)
        self.last_screenshot = screenshot

        template = Image.load(image_path)
        return screenshot.match_template(template)



import pkg_resources

__version__ = pkg_resources.get_distribution(__package__).version

from .config import Config
from .exceptions import ElementNotFoundError
from .region import Region
from .element import ElementConfig

__all__ = [Config, Region, ElementConfig, ElementNotFoundError]



import cv2
from mss import mss
import numpy

from .image import Image


class ScreenGrabber:
    def __init__(self):
        self._mss = mss()

    def grab(self, rect=None):
        if rect is None:
            monitor = self._mss.monitors[0]
        else:
            monitor = {
                "left": rect[0],
                "top": rect[1],
                "width": rect[2],
                "height": rect[3],
            }

        rgba_screenshot = numpy.array(self._mss.grab(monitor))
        rgb_screenshot = cv2.cvtColor(rgba_screenshot, cv2.COLOR_RGBA2RGB)

        expected_shape = (monitor["height"], monitor["width"])
        if rgb_screenshot.shape[:2] != expected_shape:
            return Image(
                cv2.resize(rgb_screenshot, (monitor["width"], monitor["height"]))
            )

        return Image(rgb_screenshot)



from pathlib import Path
import time
from typing import NamedTuple
import warnings

from .exceptions import ElementNotFoundError
from .locater import Locater
from .mouse import Mouse


class ElementConfig(NamedTuple):
    name: str
    click_offset: tuple = None


class Element:
    def __init__(self, element, region):
        self.boundary = region._boundary
        self.config = region._config

        self.name = element.name
        self.full_name = f"{region._name}-{element.name}"
        self.image_path = (Path(self.config.image_dir) / self.full_name).with_suffix(
            ".png"
        )
        self.click_offset = element.click_offset
        if self.click_offset is None:
            self.click_offset = (0, 0)

        self._locater = Locater()
        self._mouse = Mouse()

    def __str__(self):
        return f"<Element name='{self.name}'>"

    def click(self, timeout_seconds=None):
        if timeout_seconds is None:
            timeout_seconds = self.config.click_retry_seconds

        try:
            x, y = self._locate_centre_with_retry(timeout_seconds)
        except Exception:
            self.save_last_screenshot()
            raise

        self._mouse.click(x + self.click_offset[0], y + self.click_offset[1])

    def is_visible(self, timeout_seconds=0):
        try:
            self._locate_centre_with_retry(timeout_seconds)
            return True
        except ElementNotFoundError:
            return False

    def _locate_centre_with_retry(self, timeout_seconds):
        rect = self._locate_with_retry(timeout_seconds)
        return self._find_centre(rect)

    def _locate_with_retry(self, timeout_seconds):
        # Ensure the mouse isn't in the abort position
        if self._mouse.get_position() == (0, 0):
            self._mouse.move_to(10, 10)

        start_time = time.monotonic()
        while True:
            try:
                return self._locater.locate(self.image_path, self.boundary)
            except (ElementNotFoundError, FileNotFoundError):
                elapsed_time = time.monotonic() - start_time
                if elapsed_time > timeout_seconds:
                    raise

                # Abort if mouse pointer gets moved to (0, 0)
                if self._mouse.get_position() == (0, 0):
                    warnings.warn("Aborted - mouse pointer moved to (0, 0)")
                    raise

    def save_last_screenshot(self):
        screenshot = self._locater.last_screenshot
        if screenshot is None:
            return

        screenshot_path = (
            Path(self.config.screenshot_dir) / self.full_name
        ).with_suffix(".png")

        unique_screenshot_path = screenshot.save(screenshot_path)

        print(f"Expected Image: {self.image_path}")
        print(f"Screenshot: {unique_screenshot_path}")

    @staticmethod
    def _find_centre(rect):
        return (rect[0] + rect[2] // 2, rect[1] + rect[3] // 2)



from pathlib import Path
import inspect

from .element import Element, ElementConfig


class Region:
    def __init__(self, name, config, boundary=None):
        self._name = name
        self._config = config
        self._boundary = boundary
        self._elements = {}

        if self._config.image_dir is None:
            self._config = self._config._replace(
                image_dir=(_get_calling_dir() / "images")
            )
        if self._config.screenshot_dir is None:
            self._config = self._config._replace(
                screenshot_dir=(_get_calling_dir() / "screenshots")
            )

    def configure(self, element_configs):
        self._elements = {e.name: Element(e, self) for e in element_configs}
        return self

    def __getattr__(self, name):
        if name not in self._elements:
            self._elements[name] = Element(ElementConfig(name), self)
        return self._elements[name]


def _get_calling_dir():
    return Path(inspect.stack()[2].filename).parent



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['clickshot']

package_data = \
{'': ['*']}

install_requires = \
['mss>=4.0,<5.0',
 'numpy>=1.17,<2.0',
 'opencv-python>=4.1,<5.0',
 'pynput>=1.4,<2.0']

setup_kwargs = {
    'name': 'clickshot',
    'version': '0.3.0',
    'description': 'Easy GUI testing with pyautogui',
    'long_description': '# Clickshot\n\n[![PyPI](https://img.shields.io/pypi/v/clickshot.svg)](https://pypi.python.org/pypi/clickshot)\n[![Build Status](https://travis-ci.com/sneakypete81/clickshot.svg?branch=master)](https://travis-ci.com/sneakypete81/clickshot)\n[![Coverage](https://codecov.io/gh/sneakypete81/clickshot/graph/badge.svg)](https://codecov.io/gh/sneakypete81/clickshot)\n\nA framework for easier cross-platform GUI testing.\n\n## Installation\n\n```sh\npip install clickshot\n```\n\n## Usage\n\nTODO\n\n## Development\n\nRequires [Poetry](https://poetry.eustace.io/).\n\n```sh\ngit clone https://github.com/sneakypete81/clickshot.git\npoetry install\n```\n\nThen you can use the following:\n\n```sh\n  poetry run pytest # Run all unit tests\n  poetry run flake8 # Run the linter\n  poetry run tox    # Run all checks across all supported Python versions\n  poetry shell      # Open a venv shell with your local clone installed\n```\n',
    'author': 'sneakypete81',
    'author_email': 'sneakypete81@gmail.com',
    'url': 'https://github.com/sneakypete81/clickshot',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from typing import NamedTuple


class Config(NamedTuple):
    image_dir: str = None
    screenshot_dir: str = None

    # Default time to wait for a clickable element to appear
    click_retry_seconds: int = 30
